"""BfG Cardplay utilities."""

from bridgeobjects import Suit, SUITS, Card, SEATS, CARD_VALUES
from bfgdealer import Trick
# from .player import Player

MODULE_COLOUR = 'blue'


def get_list_of_best_scores(candidates: dict[object, int],
                            reverse: bool = False) -> list[object]:
    """Return a list of the best scoring candidates
    from a dict of candidates."""
    best_candidates = []
    (min_score, max_score) = _min_max_score(candidates, reverse)

    for key, score in candidates.items():
        if reverse:
            if score == min_score:
                best_candidates.append(key)
        else:
            if score == max_score:
                best_candidates.append(key)
    return best_candidates


def get_list_of_best_suits(player: object) -> list[object]:
    """Return a list of the best scoring candidates from a
    dict of candidates."""
    winners = {suit: len(player.winners[suit]) for suit in SUITS}
    losers = {suit: len(player.losers[suit]) for suit in SUITS}
    best_candidates = []
    scores = {}
    max_score = 99
    # for suit in SUITS:
    #     score = winners[suit] - losers[suit]
    #     ic(suit, score)
    #     scores[suit] = score
    #     if score > max_score:
    #         max_score = score

    for suit in SUITS:
        score = winners[suit] - losers[suit]
        scores[suit] = losers[suit]
        if score < max_score:
            max_score = score

    best_candidates = []
    for suit in SUITS:
        if scores[suit] == max_score:
            best_candidates.append(suit)
    return best_candidates


def _min_max_score(candidates: dict[object, int],
                   reverse: bool = False) -> int:
    """Return the maximum score from candidates."""
    max_score = 0
    min_score = 0
    for key, score in candidates.items():
        if score > max_score:
            max_score = score

    if reverse:
        min_score = max_score
        for key, score in candidates.items():
            if score < min_score:
                min_score = score
    return (min_score, max_score)


def other_suit_for_signals(suit: Suit) -> str:
    """Return the other suit for signalling."""
    if suit.name == 'S':
        other_suit = 'C'
    elif suit.name == 'C':
        other_suit = 'S'
    elif suit.name == 'H':
        other_suit = 'D'
    elif suit.name == 'D':
        other_suit = 'H'
    return other_suit


def get_suit_strength(cards: list[Card]) -> dict[str, int]:
    """Return a dict of suit high card points keyed on suit name."""
    suit_points = {suit: 0 for suit in SUITS}
    for card in cards:
        suit_points[card.suit.name] += card.high_card_points
    return suit_points


def get_seat(trick: Trick, player_index: int) -> str:
    """Return the seat of the player relative to the trick leader."""
    trick_index = SEATS.index(trick.leader)
    seat_index = (trick_index + player_index) % 4
    return SEATS[seat_index]


def highest_remaining_card(player: object, trick: Trick,
                           trumps: Suit = None) -> Card:
    """Return the highest card remaining in hand if
    player does not win trick."""

    # Ignore if 3rd player is winner:
    values = trick_card_values(trick, trumps)
    if values[2] == max(values):
        return None

    suit = trick.cards[0].suit
    winning_card = get_winning_card(trick, trumps)
    value = 2
    highest_card = None
    while value < 13 and value < winning_card.value:
        value += 1
        test_card = Card(CARD_VALUES[value], suit.name)
        unplayed_cards = player.opponents_unplayed_cards
        if test_card not in trick.cards + unplayed_cards[suit]:
            highest_card = test_card
    return highest_card


def get_winning_card(trick: Trick, trumps: Suit | None) -> Card:
    """Return the winning card in the trick."""
    card_values = trick_card_values(trick, trumps)
    for index, card in enumerate(trick.cards):
        if card_values[index] == max(card_values):
            return card
    raise ValueError(f'No winner in {trick.cards=} {trumps=}')


def trick_card_values(trick: Trick, trumps: Suit = None) -> list[int]:
    """Return a tuple of card values from a trick."""
    return card_values(trick.cards, trumps)


def card_values(cards: list[Card], trumps: Suit = None) -> list[int]:
    """Return a list of card values."""
    if len(cards) == 0:
        return [0, 0, 0, 0]

    values = []
    for index, card in enumerate(cards):
        value = _get_card_value(cards, index, trumps)
        values.append(value)
    return values


def _get_card_value(cards, index, trumps):
    """Return the trump modified value of a card."""
    card = cards[index]
    trick_suit = cards[0].suit
    value = cards[index].value
    if card.suit == trumps:
        return value + 13
    if card.suit != trick_suit:
        return 0
    return value


def play_to_unblock(hand_to_play: list[Card],
                    other_hand: list[Card]) -> Card | None:
    """Select the card to unblock the suit."""
    (long_hand, short_hand) = get_long_hand(hand_to_play, other_hand)
    if not short_hand:
        return long_hand[0]

    long_values, short_values = card_values(long_hand), card_values(short_hand)

    # Return  cards in a sequence from the short hand
    card_in_sequence = _card_in_sequence(hand_to_play, short_hand)
    if card_in_sequence:
        return card_in_sequence

    # All values in short hand are lower than corresponding one in long
    card = _all_low_in_short(hand_to_play, short_hand)
    if card:
        return card

    # The lowest card in the short hand is lower than
    # the highest card in the long hand
    if short_values[-1] < long_values[0]:
        if long_hand == hand_to_play:
            return long_hand[-1]
        return short_hand[0]

    # Default return None
    return None


def _card_in_sequence(hand_to_play: list[Card],
                      other_hand: list[Card]) -> Card | None:
    """Return top of short hand if it's in a sequence."""
    (long_hand, short_hand) = get_long_hand(hand_to_play, other_hand)
    long_values, short_values = card_values(long_hand), card_values(short_hand)
    values = sorted(long_values + short_values, reverse=True)
    sequence = _get_sequence(values)
    if len(sequence) > 1:
        suit = hand_to_play[0].suit.name
        for item in sequence:
            card = Card(CARD_VALUES[item], suit)
            if card in short_hand and short_hand == hand_to_play:
                return card
        return long_hand[-1]
    return None


def _all_low_in_short(hand_to_play: list[Card],
                      other_hand: list[Card]) -> Card | None:
    (long_hand, short_hand) = get_long_hand(hand_to_play, other_hand)
    long_values, short_values = card_values(long_hand), card_values(short_hand)

    result = True
    for index, value in enumerate(short_values):
        if value > long_values[index]:
            result = False
            break
    if result:
        return hand_to_play[0]
    return None


def _get_sequence(values):
    sequence = [values[0]]
    for index, item in enumerate(values[:-1]):
        if item == values[index + 1] + 1:
            sequence.append(values[index + 1])
        else:
            break
    return sequence


def get_long_hand(hand_1: list[Card], hand_2: list[Card]) -> tuple[list[Card]]:
    """Return a tuple of (long_hand, short_hand)."""
    long_hand, short_hand = hand_1, hand_2
    if len(hand_2) > len(hand_1):
        short_hand, long_hand = hand_1, hand_2
    return (long_hand, short_hand)


def play_lowest_winning_card(player, cards, trick):
    for card in cards[::-1]:
        if player.is_winner_declarer(card, trick):
            return card



import re

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

VERSION_FILE="bfgcardplay/_version.py"
with open(VERSION_FILE, 'r') as f_version:
    version_string = f_version.read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, version_string, re.M)
if mo:
    version_string = mo.group(1)
else:
    raise RuntimeError(f'Unable to find version string in {VERSION_FILE}.')

setup_args = dict(
    name='bfgcardplay',
    version=version_string,
    description="""
            A collection of modules that facilitate cardplay in the BfG
            environment.
        """,
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='jeff watkins',
    author_email='support@bidforgame.com',
    keywords=[(f'bridge, duplicate bridge, contract bridge, '
               f'hand, board, suit, auction, contract, card')],
    url='https://psionman@bitbucket.org/psionman/bfgcardplay.git',
    download_url='https://pypi.org/project/bfgcardplay/'
)

install_requires = [
]

if __name__ == '__main__':
    setup(
        package_data={
            'bfgcardplay': ['tests/test_data/*.pbn']
        },
        include_package_data=True,
        **setup_args,
        install_requires=install_requires
        )



"""Common functionality for cardplay."""

import random

from bridgeobjects import Suit, Card, SUITS
from bfgbidding import Hand

from .utilities import get_list_of_best_scores

MODULE_COLOUR = 'magenta'


class Cards():
    """A class representing a set of cards."""
    def __init__(self, input: list[str] | list[Card] | Hand):
        self.list = []
        self.longest_suit = None
        by_suit = {suit: [] for key, suit in SUITS.items()}
        by_suit_name = {key: [] for key, suit in SUITS.items()}
        self.by_suit = {**by_suit, **by_suit_name}
        self._value = 0
        self._create(input)

    def count(self) -> int:
        """Return the number of cards."""
        return len(self.list)

    def _create(self, input: list[str] | list[Card] | Hand):
        """Create a set of cards"""
        if isinstance(input, Hand):
            cards = input.cards
        elif not isinstance(input, list):
            raise TypeError('Cards must be a list of cards')
        else:
            cards = input

        raw_list = []
        for card in cards:
            if isinstance(card, str):
                card = Card(card)
            raw_list.append(card)
        self.list = self.sort_cards(raw_list)

        self._value = 0
        for card in self.list:
            suit = card.suit
            suit_name = suit.name
            self.by_suit[suit].append(card)
            self.by_suit[suit_name].append(card)
            self._value += card.value
        self.longest_suit = self._longest_suit()

    @staticmethod
    def sort_cards(cards: list[Card]) -> list[Card]:
        """Return a sorted list of cards."""
        return sorted(cards, reverse=True)

    def _longest_suit(self) -> Suit:
        """Return the suit with most cards."""
        suit_dict = {}
        for suit, card_list in self.by_suit.items():
            if isinstance(suit, Suit):
                suit_dict[suit] = len(card_list)
        suits = get_list_of_best_scores(suit_dict)
        suit = random.choice(suits)
        return suit

    @property
    def value(self) -> int:
        """Return the sum of the values of the cards."""
        return self._value



"""Expose packages classes."""

from bfgcardplay.src.opening_lead_suit import *
from bfgcardplay.src.opening_lead_card import *
from bfgcardplay.src.opening_lead import *
from bfgcardplay.src.suggested_card import *



from ..src.player import Player
from ..tests.utilities import get_boards
from bridgeobjects import Card


boards = get_boards('player.pbn')


def test_controls():
    board = boards['0']
    player = Player(board)
    assert player.controls == {'C': 0, 'D': 1, 'H': 2, 'S': 3}

def test_second_round_controls():
    board = boards['0']
    player = Player(board)
    assert player.second_round_controls == {'C': True, 'D': True, 'H': True, 'S': True}


def test_record_voids():
    board = boards['1']
    assert len(board.tricks) == 8
    player = Player(board)
    voids = player.voids
    assert voids['W']['S'] == False
    assert voids['W']['H'] == True
    assert voids['W']['D'] == True
    assert voids['W']['C'] == False

    assert voids['N']['S'] == False
    assert voids['N']['H'] == False
    assert voids['N']['D'] == False
    assert voids['W']['C'] == False

    assert voids['E']['S'] == False  # E is now void in S but at this stage has always followed suit

def test_short_seat():
    board = boards['2']
    player = Player(board)
    assert player.short_suit_seat == 'N'

    board = boards['3']
    player = Player(board)
    assert player.short_suit_seat == None

def test_short_hand():
    board = boards['2']
    player = Player(board)
    assert player.short_hand == board.hands['N']

    board = boards['3']
    player = Player(board)
    assert player.short_hand == None

def test_trump_cards():
    board = boards['2']
    player = Player(board)
    assert player.trump_cards == [Card("KC"), Card("TC"), Card("8C"), Card("7C")]
    assert player.opponents_trumps == [Card("AC"), Card("JC"), Card("9C"), Card("5C")]

def test_dummy_location():
    board = boards['3']
    player = Player(board)
    assert player.dummy_on_left == False
    assert player.dummy_on_right == False

    board = boards['4']
    player = Player(board)
    assert player.dummy_on_left == False
    assert player.dummy_on_right == True



"""Fourth seat card play for declarer."""
import random
from termcolor import cprint

import inspect
from ..logger import log

from bridgeobjects import SUITS, Card, Suit, CARD_VALUES
from bfgdealer import Trick
from .player import Player
from .utilities import get_list_of_best_scores, trick_card_values, get_long_hand
from .fourth_seat import FourthSeat
import bfgcardplay.src.global_variables as global_vars

MODULE_COLOUR = 'cyan'


class FourthSeatDeclarer(FourthSeat):
    def __init__(self, player: Player):
        super().__init__(player)
        self.player = player

    def selected_card(self) -> Card:
        """Return the card if the third seat."""
        player = self.player
        trick = player.board.tricks[-1]
        manager = global_vars.manager

        cards = player.cards_for_trick_suit(trick)

        # Void
        if not cards:
            return self._select_card_if_void()

        # Singleton
        if len(cards) == 1:
            return log(inspect.stack(), cards[0])
        if manager.suit_strategy == '':
            pass

        # Play low if partner is winning trick
        if self._second_player_winning_trick(cards, trick, player.trump_suit):
            return log(inspect.stack(), cards[-1])

        # When to duck - rule of 7
        if not player.trump_suit:
            if trick.suit == player.board.tricks[0].cards[0].suit:
                if self._duck_trick(player, trick, cards):
                    return log(inspect.stack(), cards[-1])
                for card in cards[::-1]:
                    if self.can_win_trick(player, card):
                        return log(inspect.stack(), card)

        # Win trick if possible
        winning_card = self._winning_card(trick)
        if winning_card:
            return winning_card

        # Play smallest card
        return log(inspect.stack(), cards[-1])

    def _select_card_if_void(self) -> Card:
        """Return card if cannot follow suit."""
        player = self.player
        trick = player.board.tricks[-1]

        # Trump if appropriate
        if player.trump_suit:
            values = trick_card_values(trick, player.trump_suit)
            if player.trump_cards:
                if values[0] > values[1] or values[2] > values[1]:
                    for card in player.trump_cards[::-1]:
                        if card.value + 13 > values[0] and card.value + 13 > values[2]:
                            return log(inspect.stack(), card)

        # Discard if winner in partner's hand
        random_suits = [suit for suit in SUITS]
        if player.trump_suit:
            random_suits.remove(player.trump_suit.name)

        # TODO do not discard an entry to partner's hand
        for suit in SUITS:
            my_cards = player.unplayed_cards[suit]
            partners_cards = player.partners_unplayed_cards[suit]
            (long_hand, short_hand) = get_long_hand(my_cards, partners_cards)
            if my_cards == short_hand:
                for index, card in enumerate(my_cards[::-1]):
                    if player.is_master_card(partners_cards[index]):
                        if card.suit.name in random_suits:
                            random_suits.remove(card.suit.name)
                            break

        random.shuffle(random_suits)
        for suit in random_suits:
            my_cards = player.unplayed_cards[suit]
            partners_cards = player.partners_unplayed_cards[suit]
            if my_cards and partners_cards:
                if len(my_cards) > 1:
                    if player.is_winner_declarer(partners_cards[0]):
                        return log(inspect.stack(), my_cards[-1])

        # Find a card that is not an honour
        for suit in random_suits:
            if player.unplayed_cards[suit]:
                if not player.unplayed_cards[suit][-1].is_honour:
                    return log(inspect.stack(), player.unplayed_cards[suit][-1])

        # Last resort - dump a low card
        for suit in random_suits:
            if player.unplayed_cards[suit]:
                return log(inspect.stack(), player.unplayed_cards[suit][-1])

        # If all else fails play a trump
        return log(inspect.stack(), player.trump_cards[-1])

    def _best_suit(self) -> Suit:
        """Select suit for signal."""
        # TODO handle no points and equal suits
        best_suit = self._strongest_suit()
        return best_suit

    def _strongest_suit(self) -> Suit | None:
        """Return the strongest_suit."""
        player = self.player
        suits = {suit: 0 for suit in SUITS}
        for suit in SUITS:
            cards = player.unplayed_cards[suit]
            for card in cards:
                suits[suit] += card.value
        if player.trump_suit:
            suits[player.trump_suit.name] = 0
        best_suits = get_list_of_best_scores(suits)
        if not best_suits:
            return player.trump_suit
        return Suit(random.choice(best_suits))

    def _duck_trick(self, player: Player, trick: Trick, cards: list[Card]) -> bool:
        """Return True if the player is to duck the trick."""
        opponents_unplayed_cards = player.opponents_unplayed_cards[trick.suit.name]
        if cards and opponents_unplayed_cards:
            # partners_cards = player.partners_suit_cards[trick.suit.name]
            # partner_can_win = False
            # if partners_cards:
            #     if  partners_cards[0].value > trick.cards[0].value:
            #         partner_can_win = True
            if (len(cards) == 2 and cards[0].value == CARD_VALUES['K'] and
                    Card('A', trick.suit.name) in opponents_unplayed_cards):
                return False
            can_win_trick = (cards[0].value > opponents_unplayed_cards[0].value and
                             cards[0].value > trick.cards[0].value and
                             cards[0].value > trick.cards[2].value)
            if self._rule_of_seven(player, trick) and can_win_trick:
                return False
        return True

    @staticmethod
    def _rule_of_seven(player: Player, trick: Trick) -> bool:
        """Return True if rule of seven applies."""
        our_cards = player.our_cards[trick.suit]
        duck_count = 7 - len(our_cards) - len(player.board.tricks)
        return duck_count < 0



from .. import next_card
from ..tests.utilities import get_boards


boards = get_boards('fourth_seat_declarer.pbn')


def test_play_partners_suit():
    """Ensure not a S in _select_card_if_void, entry to partner's hand
        for suit in SUITS:."""
    board = boards['0']
    assert next_card(board).suit.name != 'S'


def test_play_partners_suit():
    """Ensure not a S in _select_card_if_void, entry to partner's hand
        for suit in SUITS:."""
    board = boards['1']
    assert next_card(board).name != 'KH'
